@@ -1,6 +1,5 @@
 from __future__ import annotations
 
-import glob
 import json
 import os
 import platform
@@ -8,37 +7,46 @@ import shutil
 import subprocess
 import sys
 import sysconfig
-from distutils import log
-from distutils.errors import (
+import logging
+import warnings
+from setuptools.errors import (
     CompileError,
-    DistutilsExecError,
-    DistutilsFileError,
-    DistutilsPlatformError,
+    ExecError,
+    FileError,
+    PlatformError,
 )
-from distutils.sysconfig import get_config_var
+from sysconfig import get_config_var
 from pathlib import Path
-from typing import Dict, Iterable, List, NamedTuple, Optional, Set, Tuple, cast
+from typing import Dict, List, Literal, NamedTuple, Optional, Set, Tuple, cast
 
-import pkg_resources
-from semantic_version import Version
-from setuptools.command.build import build as CommandBuild  # type: ignore[import]
+from setuptools import Distribution
 from setuptools.command.build_ext import build_ext as CommandBuildExt
 from setuptools.command.build_ext import get_abi3_suffix
-from typing_extensions import Literal
+from setuptools.command.install_scripts import install_scripts as CommandInstallScripts
 
-from ._utils import format_called_process_error
+from ._utils import check_subprocess_output, format_called_process_error, Env
 from .command import RustCommand
 from .extension import Binding, RustBin, RustExtension, Strip
 from .rustc_info import (
     get_rust_host,
-    get_rust_target_list,
     get_rust_version,
     get_rustc_cfgs,
 )
 
+logger = logging.getLogger(__name__)
 
-def _check_cargo_supports_crate_type_option() -> bool:
-    version = get_rust_version()
+
+try:
+    from setuptools.command.bdist_wheel import bdist_wheel as CommandBdistWheel
+except ImportError:  # old version of setuptools
+    try:
+        from wheel.bdist_wheel import bdist_wheel as CommandBdistWheel  # type: ignore[no-redef]
+    except ImportError:
+        from setuptools import Command as CommandBdistWheel  # type: ignore[assignment]
+
+
+def _check_cargo_supports_crate_type_option(env: Optional[Env]) -> bool:
+    version = get_rust_version(env)
 
     if version is None:
         return False
@@ -70,30 +78,27 @@ class build_rust(RustCommand):
     ]
     boolean_options = ["inplace", "debug", "release", "qbuild"]
 
+    inplace: bool = False
+    debug: bool = False
+    release: bool = False
+    qbuild: bool = False
+
     plat_name: Optional[str] = None
+    build_temp: Optional[str] = None
 
     def initialize_options(self) -> None:
         super().initialize_options()
-        self.inplace = None
-        self.debug = None
-        self.release = None
-        self.qbuild = None
-        self.build_temp = None
-        self.plat_name = None
-        self.build_number = None
         self.target = os.getenv("CARGO_BUILD_TARGET")
         self.cargo = os.getenv("CARGO", "cargo")
 
     def finalize_options(self) -> None:
         super().finalize_options()
 
-        self.data_dir = self.get_data_dir()
-
-        if self.plat_name is None:
-            self.plat_name = cast(  # type: ignore[no-any-unimported]
-                CommandBuild, self.get_finalized_command("build")
-            ).plat_name
-            assert isinstance(self.plat_name, str)
+        # Inherit settings from the `build` and `build_ext` commands
+        self.set_undefined_options(
+            "build",
+            ("plat_name", "plat_name"),
+        )
 
         # Inherit settings from the `build_ext` command
         self.set_undefined_options(
@@ -103,19 +108,11 @@ class build_rust(RustCommand):
             ("inplace", "inplace"),
         )
 
-        if self.build_number is not None and not self.build_number[:1].isdigit():
-            raise ValueError("Build tag (build-number) must start with a digit.")
-
-    def get_data_dir(self) -> str:
-        components = (
-            pkg_resources.safe_name(self.distribution.get_name()).replace("-", "_"),
-            pkg_resources.safe_version(self.distribution.get_version()).replace(
-                "-", "_"
-            ),
-        )
-        if self.build_number:
-            components += (self.build_number,)
-        return "-".join(components) + ".data"
+        if self.build_temp is not None:
+            warnings.warn(
+                "`--build-temp` argument does nothing for Rust extensions, set `CARGO_TARGET_DIR` instead.",
+                DeprecationWarning,
+            )
 
     def run_for_extension(self, ext: RustExtension) -> None:
         assert self.plat_name is not None
@@ -147,19 +144,26 @@ class build_rust(RustCommand):
     def build_extension(
         self, ext: RustExtension, forced_target_triple: Optional[str] = None
     ) -> List["_BuiltModule"]:
-        target_triple = self._detect_rust_target(forced_target_triple)
-        rustc_cfgs = get_rustc_cfgs(target_triple)
+        target_triple = self._detect_rust_target(forced_target_triple, ext.env)
+        rustc_cfgs = get_rustc_cfgs(target_triple, ext.env)
 
-        env = _prepare_build_environment()
+        env = _prepare_build_environment(ext.env)
 
         if not os.path.exists(ext.path):
-            raise DistutilsFileError(
-                f"can't find Rust extension project file: {ext.path}"
+            raise FileError(
+                f"can't find manifest for Rust extension `{ext.name}` at path `{ext.path}`"
             )
 
         quiet = self.qbuild or ext.quiet
         debug = self._is_debug_build(ext)
-        use_cargo_crate_type = _check_cargo_supports_crate_type_option()
+        use_cargo_crate_type = _check_cargo_supports_crate_type_option(ext.env)
+
+        package_id = ext.metadata(quiet=quiet)["resolve"]["root"]
+        if package_id is None:
+            raise FileError(
+                f"manifest for Rust extention `{ext.name}` at path `{ext.path}` is a virtual manifest (a workspace root without a package).\n\n"
+                "If you intended to build for a workspace member, set `path` for the extension to the member's Cargo.toml file."
+            )
 
         cargo_args = self._cargo_args(
             ext=ext, target_triple=target_triple, release=not debug, quiet=quiet
@@ -191,8 +195,8 @@ class build_rust(RustCommand):
                     *ext.rustc_flags,
                 ]
 
-            # OSX requires special linker arguments
-            if rustc_cfgs.get("target_os") == "macos":
+            # Apple platforms require special linker arguments
+            if rustc_cfgs.get("target_os") in {"macos", "ios", "tvos", "watchos"}:
                 ext_basename = os.path.basename(self.get_dylib_ext_path(ext, ext.name))
                 rustc_args.extend(
                     [
@@ -212,7 +216,7 @@ class build_rust(RustCommand):
                 "wasm32",
                 "emscripten",
             ):
-                rustc_args.extend(["-C", f"link-args=-sSIDE_MODULE=2 -sWASM_BIGINT"])
+                rustc_args.extend(["-C", "link-args=-sSIDE_MODULE=2 -sWASM_BIGINT"])
 
             if use_cargo_crate_type and "--crate-type" not in cargo_args:
                 cargo_args.extend(["--crate-type", "cdylib"])
@@ -248,7 +252,7 @@ class build_rust(RustCommand):
             # If quiet, capture all output and only show it in the exception
             # If not quiet, forward all cargo output to stderr
             stderr = subprocess.PIPE if quiet else None
-            cargo_messages = subprocess.check_output(
+            cargo_messages = check_subprocess_output(
                 command,
                 env=env,
                 stderr=stderr,
@@ -260,7 +264,7 @@ class build_rust(RustCommand):
             raise CompileError(format_called_process_error(e, include_stdout=False))
 
         except OSError:
-            raise DistutilsExecError(
+            raise ExecError(
                 "Unable to execute 'cargo' - this package "
                 "requires Rust to be installed and cargo to be on the PATH"
             )
@@ -269,7 +273,6 @@ class build_rust(RustCommand):
         # it into the build directory as if it were produced by build_ext.
 
         dylib_paths = []
-        package_id = ext.metadata(quiet=quiet)["resolve"]["root"]
 
         if ext._uses_exec_binding():
             # Find artifact from cargo messages
@@ -289,7 +292,7 @@ class build_rust(RustCommand):
                         if Path(artifact).with_suffix("").name == name
                     )
                 except StopIteration:
-                    raise DistutilsExecError(
+                    raise ExecError(
                         f"Rust build failed; unable to locate executable '{name}'"
                     )
 
@@ -307,11 +310,11 @@ class build_rust(RustCommand):
                 kinds={"cdylib", "dylib"},
             )
             if len(artifacts) == 0:
-                raise DistutilsExecError(
+                raise ExecError(
                     "Rust build failed; unable to find any cdylib or dylib build artifacts"
                 )
             elif len(artifacts) > 1:
-                raise DistutilsExecError(
+                raise ExecError(
                     f"Rust build failed; expected only one cdylib or dylib build artifact but found {artifacts}"
                 )
 
@@ -329,10 +332,7 @@ class build_rust(RustCommand):
     def install_extension(
         self, ext: RustExtension, dylib_paths: List["_BuiltModule"]
     ) -> None:
-        debug_build = ext.debug if ext.debug is not None else self.inplace
-        debug_build = self.debug if self.debug is not None else debug_build
-        if self.release:
-            debug_build = False
+        debug_build = self._is_debug_build(ext)
 
         # Ask build_ext where the shared library would go if it had built it,
         # then copy it there.
@@ -346,32 +346,46 @@ class build_rust(RustCommand):
                 )
 
             if ext._uses_exec_binding():
-                ext_path = build_ext.get_ext_fullpath(module_name)
-                # remove extensions
-                ext_path, _, _ = _split_platform_and_extension(ext_path)
-
-                # Add expected extension
                 exe = sysconfig.get_config_var("EXE")
-                if exe is not None:
-                    ext_path += exe
 
-                os.makedirs(os.path.dirname(ext_path), exist_ok=True)
                 if isinstance(ext, RustBin):
-                    executable_name = module_name
+                    # will install the rust binary into the scripts directory
+                    bin_name = module_name
                     if exe is not None:
-                        executable_name += exe
-                    scripts_dir = os.path.join(
-                        build_ext.build_lib, self.data_dir, "scripts"
+                        bin_name += exe
+
+                    install_scripts = cast(
+                        CommandInstallScripts,
+                        self.get_finalized_command("install_scripts"),
                     )
-                    os.makedirs(scripts_dir, exist_ok=True)
-                    ext_path = os.path.join(scripts_dir, executable_name)
+                    ext_path = os.path.join(install_scripts.build_dir, bin_name)
                 else:
+                    # will install the rust binary into the module directory
+                    ext_path = build_ext.get_ext_fullpath(module_name)
+
+                    # add expected extension
+                    ext_path, _, _ = _split_platform_and_extension(ext_path)
+                    if exe is not None:
+                        ext_path += exe
+
+                    # if required, also generate a console script entry point
                     ext.install_script(module_name.split(".")[-1], ext_path)
             else:
+                # will install the rust library into the module directory
                 ext_path = self.get_dylib_ext_path(ext, module_name)
-                os.makedirs(os.path.dirname(ext_path), exist_ok=True)
 
-            log.info("Copying rust artifact from %s to %s", dylib_path, ext_path)
+            os.makedirs(os.path.dirname(ext_path), exist_ok=True)
+
+            # Make filenames relative to cwd where possible, to make logs and
+            # errors below a little neater
+
+            cwd = os.getcwd()
+            if dylib_path.startswith(cwd):
+                dylib_path = os.path.relpath(dylib_path, cwd)
+            if ext_path.startswith(cwd):
+                ext_path = os.path.relpath(ext_path, cwd)
+
+            logger.info("Copying rust artifact from %s to %s", dylib_path, ext_path)
 
             # We want to atomically replace any existing library file. We can't
             # just copy the new library directly on top of the old one as that
@@ -379,11 +393,18 @@ class build_rust(RustCommand):
             # This means that any process that currently uses the shared library
             # will see it modified and likely segfault.
             #
-            # We first copy the file to the same directory, as `os.rename`
+            # We first copy the file to the same directory, as `os.replace`
             # doesn't work across file system boundaries.
             temp_ext_path = ext_path + "~"
             shutil.copyfile(dylib_path, temp_ext_path)
-            os.replace(temp_ext_path, ext_path)
+            try:
+                os.replace(temp_ext_path, ext_path)
+            except PermissionError as e:
+                msg = f"{e}\n  hint: check permissions for {ext_path!r}"
+                if sys.platform == "win32":
+                    # On Windows, dll files are locked by the system when in use.
+                    msg += "\n  hint: the file may be in use by another Python process"
+                raise CompileError(msg)
 
             if sys.platform != "win32" and not debug_build:
                 args = []
@@ -396,7 +417,7 @@ class build_rust(RustCommand):
                     args.insert(0, "strip")
                     args.append(ext_path)
                     try:
-                        output = subprocess.check_output(args)
+                        check_subprocess_output(args, env=None)
                     except subprocess.CalledProcessError:
                         pass
 
@@ -447,20 +468,16 @@ class build_rust(RustCommand):
         return ext_path
 
     def _py_limited_api(self) -> _PyLimitedApi:
-        bdist_wheel = self.distribution.get_command_obj("bdist_wheel", create=False)
+        bdist_wheel = _get_bdist_wheel_cmd(self.distribution, create=False)
 
         if bdist_wheel is None:
             # wheel package is not installed, not building a limited-api wheel
             return False
         else:
-            from wheel.bdist_wheel import bdist_wheel as CommandBdistWheel
-
-            bdist_wheel_command = cast(CommandBdistWheel, bdist_wheel)  # type: ignore[no-any-unimported]
-            bdist_wheel_command.ensure_finalized()
-            return cast(_PyLimitedApi, bdist_wheel_command.py_limited_api)
+            return cast(_PyLimitedApi, bdist_wheel.py_limited_api)
 
     def _detect_rust_target(
-        self, forced_target_triple: Optional[str] = None
+        self, forced_target_triple: Optional[str], env: Env
     ) -> Optional[str]:
         assert self.plat_name is not None
         if forced_target_triple is not None:
@@ -469,14 +486,14 @@ class build_rust(RustCommand):
             return forced_target_triple
 
         # Determine local rust target which needs to be "forced" if necessary
-        local_rust_target = _adjusted_local_rust_target(self.plat_name)
+        local_rust_target = _adjusted_local_rust_target(self.plat_name, env)
 
         # Match cargo's behaviour of not using an explicit target if the
         # target we're compiling for is the host
         if (
             local_rust_target is not None
             # check for None first to avoid calling to rustc if not needed
-            and local_rust_target != get_rust_host()
+            and local_rust_target != get_rust_host(env)
         ):
             return local_rust_target
 
@@ -485,8 +502,8 @@ class build_rust(RustCommand):
     def _is_debug_build(self, ext: RustExtension) -> bool:
         if self.release:
             return False
-        elif self.debug is not None:
-            return self.debug
+        elif self.debug:
+            return True
         elif ext.debug is not None:
             return ext.debug
         else:
@@ -503,10 +520,10 @@ class build_rust(RustCommand):
         if target_triple is not None:
             args.extend(["--target", target_triple])
 
-        if release:
-            profile = ext.get_cargo_profile()
-            if not profile:
-                args.append("--release")
+        ext_profile = ext.get_cargo_profile()
+        env_profile = os.getenv("SETUPTOOLS_RUST_CARGO_PROFILE")
+        if release and not ext_profile and not env_profile:
+            args.append("--release")
 
         if quiet:
             args.append("-q")
@@ -527,6 +544,18 @@ class build_rust(RustCommand):
         if ext.args is not None:
             args.extend(ext.args)
 
+        if env_profile:
+            if ext_profile:
+                args = [p for p in args if not p.startswith("--profile=")]
+                while True:
+                    try:
+                        index = args.index("--profile")
+                        del args[index : index + 2]
+                    except ValueError:
+                        break
+
+            args.extend(["--profile", env_profile])
+
         if ext.cargo_manifest_args is not None:
             args.extend(ext.cargo_manifest_args)
 
@@ -537,7 +566,7 @@ def create_universal2_binary(output_path: str, input_paths: List[str]) -> None:
     # Try lipo first
     command = ["lipo", "-create", "-output", output_path, *input_paths]
     try:
-        subprocess.check_output(command, text=True)
+        check_subprocess_output(command, env=None, text=True)
     except subprocess.CalledProcessError as e:
         output = e.output
         raise CompileError("lipo failed with code: %d\n%s" % (e.returncode, output))
@@ -546,7 +575,7 @@ def create_universal2_binary(output_path: str, input_paths: List[str]) -> None:
         try:
             from fat_macho import FatWriter
         except ImportError:
-            raise DistutilsExecError(
+            raise ExecError(
                 "failed to locate `lipo` or import `fat_macho.FatWriter`. "
                 "Try installing with `pip install fat-macho` "
             )
@@ -580,27 +609,34 @@ def _replace_vendor_with_unknown(target: str) -> Optional[str]:
     return "-".join(components)
 
 
-def _prepare_build_environment() -> Dict[str, str]:
+def _prepare_build_environment(env: Env) -> Dict[str, str]:
     """Prepares environment variables to use when executing cargo build."""
 
+    base_executable = None
+    if os.getenv("SETUPTOOLS_RUST_PEP517_USE_BASE_PYTHON"):
+        base_executable = getattr(sys, "_base_executable")
+
+    if base_executable and os.path.exists(base_executable):
+        executable = os.path.realpath(base_executable)
+    else:
+        executable = sys.executable
+
     # Make sure that if pythonXX-sys is used, it builds against the current
     # executing python interpreter.
-    bindir = os.path.dirname(sys.executable)
+    bindir = os.path.dirname(executable)
 
-    env = os.environ.copy()
-    env.update(
+    env_vars = (env.env or os.environ).copy()
+    env_vars.update(
         {
             # disables rust's pkg-config seeking for specified packages,
             # which causes pythonXX-sys to fall back to detecting the
             # interpreter from the path.
-            "PATH": os.path.join(bindir, os.environ.get("PATH", "")),
-            "PYTHON_SYS_EXECUTABLE": os.environ.get(
-                "PYTHON_SYS_EXECUTABLE", sys.executable
-            ),
-            "PYO3_PYTHON": os.environ.get("PYO3_PYTHON", sys.executable),
+            "PATH": os.path.join(bindir, env_vars.get("PATH", "")),
+            "PYTHON_SYS_EXECUTABLE": env_vars.get("PYTHON_SYS_EXECUTABLE", executable),
+            "PYO3_PYTHON": env_vars.get("PYO3_PYTHON", executable),
         }
     )
-    return env
+    return env_vars
 
 
 def _is_py_limited_api(
@@ -643,30 +679,30 @@ def _binding_features(
                 python_version = py_limited_api[2:]
                 features.add(f"pyo3/abi3-py{python_version}")
             elif py_limited_api:
-                features.add(f"pyo3/abi3")
+                features.add("pyo3/abi3")
         return features
     elif ext.binding is Binding.RustCPython:
         return {"cpython/python3-sys", "cpython/extension-module"}
     else:
-        raise DistutilsPlatformError(f"unknown Rust binding: '{ext.binding}'")
+        raise PlatformError(f"unknown Rust binding: '{ext.binding}'")
 
 
 _PyLimitedApi = Literal["cp37", "cp38", "cp39", "cp310", "cp311", "cp312", True, False]
 
 
-def _adjusted_local_rust_target(plat_name: str) -> Optional[str]:
+def _adjusted_local_rust_target(plat_name: str, env: Env) -> Optional[str]:
     """Returns the local rust target for the given `plat_name`, if it is
     necessary to 'force' a specific target for correctness."""
 
     # If we are on a 64-bit machine, but running a 32-bit Python, then
     # we'll target a 32-bit Rust build.
     if plat_name == "win32":
-        if get_rustc_cfgs(None).get("target_env") == "gnu":
+        if get_rustc_cfgs(None, env).get("target_env") == "gnu":
             return "i686-pc-windows-gnu"
         else:
             return "i686-pc-windows-msvc"
     elif plat_name == "win-amd64":
-        if get_rustc_cfgs(None).get("target_env") == "gnu":
+        if get_rustc_cfgs(None, env).get("target_env") == "gnu":
             return "x86_64-pc-windows-gnu"
         else:
             return "x86_64-pc-windows-msvc"
@@ -752,3 +788,14 @@ def _replace_cross_target_dir(path: str, ext: RustExtension, *, quiet: bool) ->
     cross_target_dir = ext._metadata(cargo="cross", quiet=quiet)["target_directory"]
     local_target_dir = ext._metadata(cargo="cargo", quiet=quiet)["target_directory"]
     return path.replace(cross_target_dir, local_target_dir)
+
+
+def _get_bdist_wheel_cmd(
+    dist: Distribution, create: Literal[True, False] = True
+) -> Optional[CommandBdistWheel]:
+    try:
+        cmd_obj = dist.get_command_obj("bdist_wheel", create=create)
+        cmd_obj.ensure_finalized()  # type: ignore[union-attr]
+        return cast(CommandBdistWheel, cmd_obj)
+    except Exception:
+        return None
